#pragma once
#include <boost/uuid/uuid.hpp>
#include <string>
#include <kr2_program_api/api_v1/program_interface.h>
#include "domain/physical_quantities.h"
#include "robot/robot_motion.h"

namespace spin
{
    constexpr double bitXOffsetM = 0.0355;
    constexpr double bitYOffsetM = 0.0195;
    constexpr double bitZOffsetM = 0.0075;
    constexpr double inFrontOfBitYOffsetM = -0.010;
    constexpr double endSafeYOffsetM = 0.005;

    constexpr int toolSpeed_rpm = 100;
    constexpr double robotSpeed_m_s = 0.350;
    constexpr double safeRobotSpeed_m_s = 0.100;
    constexpr double attachBitSpeed_m_s = 0.010;
    constexpr double startSafeAboveOffset_m = 0.07;
    constexpr double endSafeAboveOffset_m = 0.025;

    class BitChangingStation
    {
    public:
        enum class Bit
        {
            BIT_1,
            BIT_2,
            BIT_3
        };
        enum class Function
        {
            ATTACH,
            DETACH
        };
        BitChangingStation(const kr2_program_api::RobotPose &position1,
                           const kr2_program_api::RobotPose &position2,
                           const kr2_program_api::RobotPose &position3,
                           std::shared_ptr<RobotMotion> robotMotion,
                           std::shared_ptr<IToolController> toolController);
        void attachBit(const Bit &bit, const bool twistingMotion) const;
        void detachBit(const Bit &bit) const;
        kr2_program_api::RobotPose referencePose() const { return referencePose_; }

        class TeachPoints
        {
        public:
            static TeachPoints createAndVerifyFromUnorderedPositions(const kr2rc_api::Vector &position1, const kr2rc_api::Vector &position2, const kr2rc_api::Vector &position3);
            const kr2rc_api::Vector pointA_;
            const kr2rc_api::Vector pointB_;
            const kr2rc_api::Vector pointC_;

        private:
            TeachPoints(const kr2rc_api::Vector &pointA, const kr2rc_api::Vector &pointB, const kr2rc_api::Vector &pointC);
        };

    private:
        void moveStepwiseWithMotorChangingDirection(const kr2_program_api::RobotPose &from, const Length &distance, const Length &stepLength,  const Speed &robotSpeed, const int motorSpeed_rpm) const;
        kr2_program_api::RobotPose poseOf(const Bit &bit) const;
        kr2_program_api::RobotPose poseInFrontOf(const Bit &bit) const;
        kr2_program_api::RobotPose poseEndSafe(const Bit &bit) const;
        kr2_program_api::RobotPose poseOfPointB() const;
        kr2_program_api::RobotPose poseAbovePointB() const;
        kr2rc_api::Vector calculateReferencePosition(BitChangingStation::TeachPoints &teachPoints);
        kr2rc_api::Rotation calculateReferenceRotation(BitChangingStation::TeachPoints &teachPoints);
        void calculateReferencePose(const kr2_program_api::RobotPose &position1, const kr2_program_api::RobotPose &position2, const kr2_program_api::RobotPose &position3);
        kr2_program_api::RobotPose referencePose_;
        std::shared_ptr<RobotMotion> robotMotion_;
        std::shared_ptr<IToolController> toolController_;
    };

} // namespace spin
